﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Alarmeinheit                                                                        //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2024 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//   Topologie des Alarmsystems:                                                        //
//                                                                                      //
//                                 +-- Cy4Device ---+                                   //
//                                 |                |                                   //
//                                 |  AlarmControl  |                                   //
//                                 |                |                                   //
//                                 +-------+--------+                                   //
//                                         |                                            //
//                                         |                                            //
//    ··· ---+--- Cy4NET --------+---------+---------+--------- ··· -------+--- ···     //
//           |                   |                   |                     |            //
//           |                   |                   |                     |            //
//    +- Cy4Device -+     +- Cy4Device -+     +- Cy4Device -+       +- Cy4Device -+     //
//    |             |     |             |     |             |       |             |     //
//    |  AlarmUnit  |     |  AlarmUnit  |     |  AlarmUnit  |  ...  |  AlarmUnit  |     //
//    |             |     |             |     |             |       |             |     //
//    +-------------+     +-------------+     +-------------+       +-------------+     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "AlarmUnit.h"
#include "Cy4Net.h"
#include "Timer.h"
#include "Storing.h"
#include "TextFeed.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                         // Aktions-Codes des Alarmsystems für:
{   ACTION_CODE_REQUEST_STATE = 99,  // · Rundruf-Aufforderung an die Alarm-Steuereinheit
    ACTION_CODE_DEACTIVATION  =  1,  // · Deaktivierung des Alarmsystems
    ACTION_CODE_ACTIVATION	  =  3,  // · Aktivierung des Alarmsystems
    ACTION_CODE_TRIGGER_ALARM =  5   // · Alarmauslösung
} ActionCode;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static AlarmUnitState  state                = ALARMUNIT_STATE_STARTED;
static UInt08          parameter            = 0xFF;
static Timer           stateMachineTimer    = TIMER_SECONDS(0);
static bool            stateReceived        = false;
static UInt08          announcementDuration = 0;
static const char*     eepTriggerTextFeed   = NULL;

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichten-Ausstatter  //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupRequestStateBroadcastMessage(Cy4Message* message)
{
    Cy4Address address = CY4ADDRESS_Create(CY4NET_Address().net, 0);
    UInt08 data[3] = { 0x20, 0xFF, (UInt08)ACTION_CODE_REQUEST_STATE };
    CY4MESSAGE_SetupWrite(message, address, 0xE02010, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupDeactivateAlarmSystemMessage(Cy4Message* message)
{
    Cy4Address address = CY4ADDRESS_Create(CY4NET_Address().net, 0);
    UInt08 data[3] = { 0x20, 0xFF, (UInt08)ACTION_CODE_DEACTIVATION };	
    CY4MESSAGE_SetupWrite(message, address, 0xE02020, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupActivateAlarmSystemMessage(Cy4Message* message)
{
    Cy4Address address = CY4ADDRESS_Create(CY4NET_Address().net, 0);
	UInt08 data[4] = { 0x20, 0xFF, (UInt08)ACTION_CODE_ACTIVATION, announcementDuration };
    CY4MESSAGE_SetupWrite(message, address, 0xE02030, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupTriggerAlarmSystemMessage(Cy4Message* message)
{
    Cy4Address address = CY4ADDRESS_Create(CY4NET_Address().net, 0);
    UInt08 data[55] = { 0x20, 0xFF, (UInt08)ACTION_CODE_TRIGGER_ALARM, announcementDuration, 0 };
    UInt08 count = 4;
    if (eepTriggerTextFeed != NULL)
    {   TextFeed textFeed;
        STORING_LoadBlock(eepTriggerTextFeed, textFeed.text, sizeof(textFeed.text));
        TEXTFEED_Serialize(&textFeed, &data[4]);
        count += TEXTFEED_Length(&textFeed)+1;
    }
    CY4MESSAGE_SetupWrite(message, address, 0xE02040, data, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfrage-Bearbeiter  ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ALARMUNIT_StateBroadcast (write only)                                       //
//                                                                                      //
//  Beschreibung: Rundruf über den die Alarm-Steuereinheit den Teilnehmern des Alarm-   //
//                Verbundsystems den aktuellen Zustand mitteilt. [PAR] enthält zu-      //
//                standsspezifische Parameter oder FFh wenn ungenutzt oder unbeschrie-  //
//                ben.                                                                  //
//                Die Schreibaufforderung ist nur gültig, wenn [TRx] den unten ange-    //
//                gebene Auslösewert mitführt.                                          //
//                                                                                      //
//              0   1   2   3                                                           //
//            +---+---+---+---+                                                         //
//  Belegung: |TR1|TR2|STA|PAR|                                                         //
//            +---+---+---+---+                                                         //
//                                                                                      //
//  [TRx] Auslösewert: 20FFh (8447).                                                    //
//  [STA] Zustand des Alarm-Verbundsystems: 1 = Alarmsystem abgeschaltet                //
//                                          2 = Alarmsystemaktivierung angekündigt      //
//                                          3 = Alarmsystem aktiviert                   //
//                                          4 = Alarmauslösung angekündigt              //
//                                          5 = Alarm ausgelöst                         //
//                                          x = wird ignoriert                          //
//  [PAR] Parameter zum Zustand: 1: FFh (ungenutzt)                                     //
//                               2: Dauer bis zur Aktivierung: 0 .. 240 [s], (FFh).     //
//                               3: FFh (ungenutzt)                                     //
//                               4: Dauer bis zur Alarmauslösung: 0 .. 240 [s], (FFh).  //
//                               5: Dauer bis zur Alarmabschaltung: 1 .. 60 min, (FFh). //
//                               x: FFh (ungenutzt)                                     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
void
OnALARMUNIT_StateBroadcast(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[4] = { 0, 0, 0, 0xFF };
    CY4SERVICE_TransferPayload(service, message, data);
    if (ToWord(data[0], data[1]) == 0x20FF)
    {   if (data[2] > 0 && data[2] < 6)
        {   state = (AlarmUnitState)data[2];
            parameter = data[3];
            stateReceived = true;
        }
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
RequestStateBroadcast(void)
{
    // Aufforderung an die Alarm-Steuereinheit, den aktuellen Status als Rundruf zu senden
    if (!CY4NET_IsPending(SetupRequestStateBroadcastMessage, NULL))
    {   CY4NET_Post(SetupRequestStateBroadcastMessage, NULL);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Zustandsautomat: Handhabung des Alarmsystem-Zustandes (& Anfrage)                   //
//                                                                                      //
//  Beschreibung: Nach Systemstart befindet sich das lokale Alarmsystem (das Alarm-     //
//                system dieses Gerätes) im Zustand 'ALARM_UNIT_STATE_STARTED'. Dabei   //
//                wartet der Automat auf die Rundmitteilung seitens der Alarmsteuer-	//
//                einheit, die den Mitgliedern des Alarmsystems-Verbundes den offi-		//
//                ziellen Alarmsystem-Zustand mitteilt. Bleibt diese Benachrichtigung   //
//                länger als 'CFG_ALARM_UNIT_STARTUP_DELAY' Sekunden aus, wird der		//
//                Automat aktiv und fordert die Alarmsysteuereinehit dazu auf, die		//
//                Zustands-Rundmitteilung zu versenden. Dafür ist ein Zeitrahmen von	//
//                'CFG_ALARM_UNIT_STATE_RECEIVE_TIME_SLOT' Sekunden vorgesehen. Geht	//
//                auch innerhalb dieser Zeitspanne keine Zustands-Rundmitteilung ein,   //
//                wird das lokale Alarmsystem in den Zustand 'ALARM_UNIT_STATE_UNKNOWN' //
//                versetzt. Von dort aus wird die Aufforderung der Alarmsteuereinheit	//
//                im Invervall von 'CFG_ALARM_UNIT_STATE_RETRY_INTERVAL' Sekunden		//
//				  wiederholt.															//
//                Ein übermittelter Alarmsystem-Zustand muss spätestens nach 'CFG_-     //
//                ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT' Minuten erneut bestätigt       //
//                werden. Andernfalls wird das lokale Alarmsystem wieder in den Zustand //
//                'ALARM_UNIT_STATE_UNKNOWN' versetzt.                                  //
//                                                                                      //
//                                                                                      //
//                (O)                                                                   //
//                 | StartTimer(CFG_ALARM_UNIT_STARTUP_DELAY)                           //
//                 |                                                                    //
//                \|/                                                                   //
//         +---------------+ EventStateReceived                                       	//
//         |               | -> StartTimer(CFG_ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT)   //
//         |    STARTED    +---------------------------------------------------------+  //
//         |               |                                                         |  //
//         +-------+-------+                                                         |  //
//                 | EventTimeout                                                    |  //
//                 | -> RequestStateBroadcast()                                      |  //
//                 | -> StartTimer(CFG_ALARM_UNIT_STATE_RECEIVE_TIME_SLOT)           |  //
//                 |                                                                 |  //
//                \|/                                                                |  //
//          +-------------+ EventStateReceived                                       |  //
//          |             | -> StartTimer(CFG_ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT) |  //
//   +----->|    AWAIT    +--------------------------------------------------------->|  //
//   |      |             |                                                          |  //
//   |      +------+------+                                                          |  //
//   |             | EventTimeout                                                    |  //
//   |             | -> StartTimer(CFG_ALARM_UNIT_STATE_RETRY_INTERVAL)              |  //
//   |             |                                                                 |  //
//   |            \|/                                                                |  //
//   |      +------------+ EventStateReceived                                        |  //
//   |      |            | -> StartTimer(CFG_ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT)  |  //
//   |      |    DEAD    +---------------------------------------------------------->|  //
//   |      |            |                                                           |  //
//   |      +------+-----+                                                           |  //
//   |             | EventTimeout                                                    |  //
//   |             | -> RequestStateBroadcast()                                      |  //
//   |             | -> StartTimer(CFG_ALARM_UNIT_STATE_RECEIVE_TIME_SLOT)           |  //
//   |             |                                                                 |  //
//   |<------------+                                                                 |  //
//   |                                                                               |  //
//   | EventTimeout                                                                  |  //
//   | -> RequestStateBroadcast()                                                    |  //
//   | -> StartTimer(CFG_ALARM_UNIT_STATE_RECEIVE_TIME_SLOT)                         |  //
//   |      +-------------+                                                          |  //
//   |      |             |                                                          |  //
//   +------+    ALIVE    |<---------------------------------------------------------+  //
//          |             |                                                  /|\        //
//          +------+------+                                                   |         //
//                 | EventStateReceived                                       |         //
//                 | -> StartTimer(CFG_ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT) |         //
//                 +----------------------------------------------------------+         //
//                                                                                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedStateHandling(void)
{
    typedef enum    // Zustände der Alarm-Steuereinheit-Kontaktierung:
    {   STARTED,    // · System wurde gestartet
        AWAIT,      // · Warten auf Reaktion der Alarm-Steuereinheit
        ALIVE,      // · Alarm-Steuereinheit hat reagiert
        DEAD        // · Alarm-Steuereinheit reagiert nicht
    } PeerState;

    static PeerState peerState = STARTED;

    if (stateReceived)
    {   // Alarm-Steuereinheit hat reagiert und den Alarmsystem-Zustand übermittelt
        stateReceived = false;
        TIMER_Start(&stateMachineTimer, CFG_ALARM_UNIT_STATE_CONFIRMATION_TIMEOUT*60);
        peerState = ALIVE;
    }
    else if (TIMER_HasExpired(&stateMachineTimer))
    {   if (peerState == STARTED || peerState == DEAD || peerState == ALIVE)
        {   // Expliziter Kontaktversuch mit der Alarm-Steuereinheit
            RequestStateBroadcast();
            TIMER_Start(&stateMachineTimer, CFG_ALARM_UNIT_STATE_RECEIVE_TIME_SLOT);
            peerState = AWAIT;
        }
        else if (peerState == AWAIT)
        {   // Alarm-Steuereinheit reagiert nicht, Kontaktierung später erneut versuchen
            TIMER_Start(&stateMachineTimer, CFG_ALARM_UNIT_STATE_RETRY_INTERVAL);
            peerState = DEAD;
            state = ALARMUNIT_STATE_UNKNOWN;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
AlarmUnitState
ALARMUNIT_State(void)
{
    return state;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Activate(UInt08 duration)
{
	announcementDuration = duration;
	if (!CY4NET_IsPending(SetupActivateAlarmSystemMessage, NULL))
	{   CY4NET_Post(SetupActivateAlarmSystemMessage, NULL);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Deactivate(void)
{    
    if (!CY4NET_IsPending(SetupDeactivateAlarmSystemMessage, NULL))
    {   CY4NET_Post(SetupDeactivateAlarmSystemMessage, NULL);    
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Toggle(UInt08 duration)
{
    if (state == ALARMUNIT_STATE_DISABLED) ALARMUNIT_Activate(duration);
    else                                   ALARMUNIT_Deactivate();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Trigger(UInt08 duration, const char* eepText)
{
    announcementDuration = duration;
    eepTriggerTextFeed   = eepText;
    if (state == ALARMUNIT_STATE_ACTIVATED)
	{	if (!CY4NET_IsPending(SetupTriggerAlarmSystemMessage, NULL))
        {   CY4NET_Post(SetupTriggerAlarmSystemMessage, NULL);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Startup(void)
{
    TIMER_Start(&stateMachineTimer, CFG_ALARM_UNIT_STARTUP_DELAY);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Proceed(void)
{
    ProceedStateHandling();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ALARMUNIT_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
