﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2024 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"
#include "AlarmUnit.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                 // Kontaktzustand:
{   CONTACT_STATE_CLOSED,    // · Kontakt ist geschlossen
    CONTACT_STATE_OPEN,      // · Kontakt ist geöffnet
    CONTACT_STATE_UNKNOWN    // · Kontaktzustand unbekannt (Anfangszustand)
} ContactState;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static ContactState  contactState[CFG_CONTACT_COUNT] = { CONTACT_STATE_UNKNOWN,
                                                         CONTACT_STATE_UNKNOWN,
                                                         CONTACT_STATE_UNKNOWN
                                                       };

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SerializeEventSignal(const EventSignal* signal, UInt08 data[3])
{
    data[0] = (UInt08)signal->toneSignal;
    data[1] = (UInt08)signal->lightSignal;
    data[2] = (UInt08)signal->brightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
static EventSignal
DeserializeEventSignal(const UInt08 data[3])
{
    EventSignal signal;
    if (data[0] < (UInt08)TONEEVENTSIGNAL_NUMBER) signal.toneSignal = (ToneEventSignal)data[0];
    else                                          signal.toneSignal = TONEEVENTSIGNAL_NONE;
    if (data[1] < (UInt08)LIGHTEVENTSIGNAL_NUMBER) signal.lightSignal = (LightEventSignal)data[1];
    else                                           signal.lightSignal = LIGHTEVENTSIGNAL_NONE;
    signal.brightness = Bound(10, data[2], 100);
    return signal;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
SerializeStateSignal(const StateSignal* signal, UInt08 data[3])
{
    data[0] = (UInt08)signal->toneSignal;
    data[1] = (UInt08)signal->lightSignal;
    data[2] = (UInt08)signal->brightness;
}

//////////////////////////////////////////////////////////////////////////////////////////
static StateSignal
DeserializeStateSignal(const UInt08 data[3])
{
    StateSignal signal;

    if (data[0] < (UInt08)TONESTATESIGNAL_NUMBER)  signal.toneSignal  = (ToneStateSignal)data[0];
    else                                           signal.toneSignal  = TONESTATESIGNAL_OFF;
    if (data[1] < (UInt08)LIGHTSTATESIGNAL_NUMBER) signal.lightSignal = (LightStateSignal)data[1];
    else                                           signal.lightSignal = LIGHTSTATESIGNAL_OFF;
    signal.brightness = Bound(10, data[2], 100);
    return signal;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfrage-Bearbeiter  ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactState (read only)                                                    //
//                                                                                      //
//  Beschreibung: Informiert über den Zustand der Kontakte unter Brücksichtigung der    //
//  Ereignis-Verzögerungen (s. ContactEventConfiguration.xED).                          //
//                                                                                      //
//              0   1   2                                                               //
//            +---+---+---+                                                             //
//  Belegung: |C1S|C2S|C3S|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [CnS] Zustand Kontakt n: 0|1 = geöffnet|geschlossen.                                //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactState(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[CFG_CONTACT_COUNT];

    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   bool closed = contactState[i] == CONTACT_STATE_CLOSED;
        data[i] = closed;
    }
    CY4SERVICE_TransferPayload(service, message, data);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventConfiguration (read write)                                      //
//                                                                                      //
//  Beschreibung: Bestimmt Verhalten und Aktivität des jeweiligen Kontaktwechsels im    //
//                Alarmsystem. ContactEventConfiguration ist ein Gemeinschaftsbear-     //
//                beiter folgender Dienste:                                             //
//                                                                                      //
//                  130100h:5: Konfiguration für Ereignis Kontakt 1 geschlossen         //
//                  130180h:5: Konfiguration für Ereignis Kontakt 1 geöffnet            //
//                  130200h:5: Konfiguration für Ereignis Kontakt 2 geschlossen         //
//                  130280h:5: Konfiguration für Ereignis Kontakt 2 geöffnet            //
//                  130300h:5: Konfiguration für Ereignis Kontakt 3 geschlossen         //
//                  130380h:5: Konfiguration für Ereignis Kontakt 3 geöffnet            //
//                                                                                      //
//                Signalisierung von Konaktereignissen findet nur statt, wenn die       //
//                Alarmeinheit deaktiviert ist.                                         //
//                                                                                      //
//              0    1  2   3   4                                                       //
//            +---+---+---+---+---+                                                     //
//  Belegung: |DUR|TSE|LSE|BRT|CRC|                                                     //
//            +---+---+---+---+---+                                                     //
//                                                                                      //
//  [DUR] Zustandsdauer bis o.g. Ereignis eintritt: 0 .. 240 [s].                       //
//  [TSE] Akustisches Signal des Piezosignalgebers bei o.g. Ereignis und deaktivierten  //
//        Alarmsystem:                                                                  //
//           0 = Lautlos (Voreinstellung).                                              //
//           1 = Tief Hoch                                                              //
//           2 = Hoch Tief                                                              //
//           3 = Tief Mitte Hoch                                                        //
//           4 = Hoch Mitte Tief                                                        //
//           5 = Taste gedrückt                                                         //
//           6 = Oktave2Hoch                                                            //
//           7 = Doppelton                                                              //
//           8 = Trippelton                                                             //
//           9 = Zwitscher                                                              //
//          10 = Kurzer Piepton C8                                                      //
//          11 = Langer Piepton C8                                                      //
//          12 = Kurzer Piepton C6                                                      //
//          13 = Langer Piepton C6                                                      //
//           * = Lautlos                                                                //
//  [LSE] Lichtsignaleffekt der LED bei o.g. Ereignis und deaktivierten Alarmsystem:    //
//           0 = Aus (Voreinstellung).                                                  //
//           1 = Blinkt einmal                                                          //
//           2 = Blinkt zweimal                                                         //
//           3 = Blinkt dreimal                                                         //
//           4 = Blinkt sechsmal                                                        //
//           5 = Blitzt einmal                                                          //
//           6 = Blitzt zweimal                                                         //
//           7 = Blitzt dreimal                                                         //
//           8 = Blitzt sechsmal                                                        //
//           9 = Einblenden                                                             //
//          10 = Ausblenden                                                             //
//          11 = Flackern                                                               //
//           * = Aus                                                                    //
//  [BRT] Helligkeit der LED beim Lichtsignaleffekt [LSE]: 10 .. 100 %                  //
//  [CRC] Aufgabe eines Kontaktes im Alarmsystem:                                       //
//           0 = Kontakt hat keine Aufgabe (voreinstellung)                             //
//           1 = Oben genanntes Kontaktereignis aktiviert   Alarmsystem                 //
//           2 = Oben genanntes Kontaktereignis deaktiviert Alarmsystem                 //
//           3 = Oben genanntes Kontaktereignis aktiviert und deaktiviert Alarmsystem   //
//           4 = Oben genanntes Kontaktereignis löst Alarm aus                          //
//           * = Kontakt hat keine Aufgabe                                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventConfiguration(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 contactIndex;    // 0,1 = Kontakt 1,2
    UInt08 openEvent;       // 0,1 = geschlossen, geöffnet
    if      (location < 0x130180) { contactIndex = 0; openEvent = 0; }
    else if (location < 0x130200) { contactIndex = 0; openEvent = 1; }
    else if (location < 0x130280) { contactIndex = 1; openEvent = 0; }
    else if (location < 0x130300) { contactIndex = 1; openEvent = 1; }
    else if (location < 0x130380) { contactIndex = 2; openEvent = 0; }
    else                          { contactIndex = 2; openEvent = 1; }
    UInt08 data[5];
    {   ContactEventConfiguration configuration;
        STORING_Load(STO_ContactEventConfiguration[contactIndex][openEvent], &configuration);
        data[0] = (UInt08)configuration.stateDurationUntilEvent;
        SerializeEventSignal(&configuration.eventSignal, &data[1]);
        data[4] = (UInt08)configuration.alarmRole;
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   ContactEventConfiguration configuration;
        configuration.stateDurationUntilEvent = Bound(0, data[0], 240);
        configuration.eventSignal = DeserializeEventSignal(&data[1]);
        if (data[4] < 5) configuration.alarmRole = (ContactAlarmRole)data[4];
        else             configuration.alarmRole = CONTACT_ALARM_ROLE_NONE;
        STORING_Save(&configuration, STO_ContactEventConfiguration[contactIndex][openEvent]);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: ContactEventSignalTest (write only)                                         //
//                                                                                      //
//  Beschreibung: Löst zu Probezwecken die ausgewählte Ereignis-Signalisierung aus      //
//                (Ton- und Lichtsignal). ContactEventSignalTest ist ein Gemeinschafts- //
//                bearbeiter folgender Dienste:                                         //
//                                                                                      //
//                  13017Ah:1: Signaltest für Ereignis Kontakt 1 geschlossen            //
//                  1301FAh:1: Signaltest für Ereignis Kontakt 1 geöffnet               //
//                  13027Ah:1: Signaltest für Ereignis Kontakt 2 geschlossen            //
//                  1302FAh:1: Signaltest für Ereignis Kontakt 2 geöffnet               //
//                  13037Ah:1: Signaltest für Ereignis Kontakt 3 geschlossen            //
//                  1303FAh:1: Signaltest für Ereignis Kontakt 3 geöffnet               //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |TRG|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//                                                                                      //
//  [TRG] Auslöser der Signalisierung des o.g. Ereignisses: 94 (5Eh, '^').              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnContactEventSignalTest(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 contactIndex;    // 0 = Kontakt 1, 1 = Kontakt 2
    UInt08 openEvent;       // 0 = geschlossen, 1 = geöffnet
    if      (location < 0x1301FA) { contactIndex = 0; openEvent = 0; }
    else if (location < 0x13027A) { contactIndex = 0; openEvent = 1; }
    else if (location < 0x1302FA) { contactIndex = 1; openEvent = 0; }
    else if (location < 0x13037A) { contactIndex = 1; openEvent = 1; }
    else if (location < 0x1303FA) { contactIndex = 2; openEvent = 0; }
    else                          { contactIndex = 2; openEvent = 1; }
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E)
    {   EventSignal signal;
        STORING_Load(STO_ContactEventConfiguration[contactIndex][openEvent].eventSignal, &signal);
        LIGHTSIGNAL_Play(signal.lightSignal, signal.brightness);
        TONESIGNAL_Play(signal.toneSignal);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmAnnouncementDuration (read write)                                      //
//                                                                                      //
//  Beschreibung: Zugriff auf die Dauer der Ankündigungen, die die Anwendung bei Akti-  //
//                vierung des Alarmsystems bzw. bei Auslösung eines Alarmes mitliefert. //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |AAD|TAD|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [AAD] Dauer der Aktivierungsankündigung: 0 .. 240 [s].                              //
//  [TAD] Dauer der Alarm-Auslöseankündigung: 0 .. 240 [s].                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmAnnouncementDuration(const Cy4Service* service, Cy4Message* message)
{
    UInt08 data[2];
    STORING_Load(STO_ActivationAnnouncementDuration, &data[0]);
    STORING_Load(STO_TriggerAnnouncementDuration, &data[1]);
    CY4SERVICE_TransferPayload(service, message, data);
    data[0] = Bound(0, data[0], 240);
    data[1] = Bound(0, data[1], 240);
    STORING_Save(&data[0], STO_ActivationAnnouncementDuration);
    STORING_Save(&data[1], STO_TriggerAnnouncementDuration);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmTriggerTextFeed (read write)                                           //
//                                                                                      //
//  Beschreibung: Zugriff auf den Text der Rundmitteilung, der bei Auslösung eines      //
//                Alarmes seitens der Alarm-Steuereinheit verteilt wird. Der Text wird  //
//                wird von der Applikation bei Auslösung eines Alarmes mit übergeben.   //
//                                                                                      //
//              0   1  ...  49  50                                                      //
//            +---+---+   -+---+---+                                                    //
//  Belegung: |T01|T02|... |T50| 0 |                                                    //
//            +---+---+   -+---+---+                                                    //
//                                                                                      //
//  [Txx] Zeichen der Rundmitteilung bei o.g. Ereignis: 0 .. 255                        //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmTriggerTextFeed(const Cy4Service* service, Cy4Message* message)
{
    TextFeed textFeed;
    STORING_Load(STO_TriggerTextFeed, &textFeed);
    UInt08 data[51];
    TEXTFEED_Serialize(&textFeed, data);
    CY4SERVICE_TransferPayload(service, message, data);
    TEXTFEED_Deserialize(data, &textFeed);
    STORING_Save(&textFeed, STO_TriggerTextFeed);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmUnitState (read only)                                                  //
//                                                                                      //
//  Beschreibung: Liefert den aktuellen Zustand der Alarmeinheit. Mit Ausnahme des      //
//                Startzustandes (0) erhält die Alarmeinheit den aktuellen Zustand via  //
//                Rundruf durch die Alarm-Steuereinheit.                                //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |STA|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [STA] Aktueller Zustand der Alarmeinheit:                                           //
//          0 = Alarmeinheit gestartet                                                  //
//          1 = Alarmeinheit deaktiviert                                                //
//          2 = Aktivierung der Alarmeinheit angekündigt                                //
//          3 = Alarmeinheit aktiviert                                                  //
//          4 = Alarmauslösung angekündigt                                              //
//          5 = Alarm ausgelöst                                                         //
//          6 = Zustand unbekannt                                                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmUnitState(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    CY4MESSAGE_Data(message)[0] = (UInt08)ALARMUNIT_State();
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmEventSignal (read write)                                               //
//                                                                                      //
//  Beschreibung: Legt die Signalisierung bei bestimmten Alarmereignissen fest. Alarm-  //
//                EventSignal ist Gemeinschaftsbearbeiter folgender Dienste:            //
//                                                                                      //
//          401000h:3: Signale bei Ereignis: Alarmeinheit abgeschaltet                  //
//          401100h:3: Signale bei Ereignis: Aktivierung der Alarmeinheit angekündigt   //
//          401200h:3: Signale bei Ereignis: Alarmeinheit aktiviert                     //
//          401300h:3: Signale bei Ereignis: Alarmauslösung angekündigt                 //
//          401400h:3: Signale bei Ereignis: Alarm ausgelöst                            //
//                                                                                      //
//              0    1   2                                                              //
//            +---+---+---+                                                             //
//  Belegung: |TSE|LSE|BRT|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [TSE] Akustisches Signal des Piezosignalgebers bei o.g. Ereignis:                   //
//           0 = Lautlos (Voreinstellung).                                              //
//           1 = Tief Hoch                                                              //
//           2 = Hoch Tief                                                              //
//           3 = Tief Mitte Hoch                                                        //
//           4 = Hoch Mitte Tief                                                        //
//           5 = Taste gedrückt                                                         //
//           6 = Oktave2Hoch                                                            //
//           7 = Doppelton                                                              //
//           8 = Trippelton                                                             //
//           9 = Zwitscher                                                              //
//          10 = Kurzer Piepton C8                                                      //
//          11 = Langer Piepton C8                                                      //
//          12 = Kurzer Piepton C6                                                      //
//          13 = Langer Piepton C6                                                      //
//           * = Lautlos                                                                //
//  [LSE] Lichtsignaleffekt der LED bei o.g. Ereignis:                                  //
//           0 = Aus (Voreinstellung).                                                  //
//           1 = Blinkt einmal                                                          //
//           2 = Blinkt zweimal                                                         //
//           3 = Blinkt dreimal                                                         //
//           4 = Blinkt sechsmal                                                        //
//           5 = Blitzt einmal                                                          //
//           6 = Blitzt zweimal                                                         //
//           7 = Blitzt dreimal                                                         //
//           8 = Blitzt sechsmal                                                        //
//           9 = Einblenden                                                             //
//          10 = Ausblenden                                                             //
//          11 = Flackern                                                               //
//           * = Aus                                                                    //
//  [BRT] Helligkeit der LED beim Lichtsignaleffekt [LSE]: 10 .. 100 %                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmEventSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 index;
    if     ( location < 0x401100 ) index = 0;
    else if( location < 0x401200 ) index = 1;
    else if( location < 0x401300 ) index = 2;
    else if( location < 0x401400 ) index = 3;
    else                           index = 4;
    EventSignal signal;
    STORING_Load(STO_EventSignal[index], &signal);
    UInt08  data[3];
    SerializeEventSignal(&signal, data);
    CY4SERVICE_TransferPayload(service, message, data);
    signal = DeserializeEventSignal(data);
    STORING_Save(&signal, STO_EventSignal[index]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmEventSignalTest (write only)                                           //
//                                                                                      //
//  Beschreibung: Löst zu Probezwecken die ausgewählte Ereignis-Signalisierung aus      //
//                (Ton- und Lichtsignal). AlarmEventSignalTest ist ein Gemeinschafts-   //
//                bearbeiter folgender Dienste:                                         //
//                                                                                      //
//       401F00h:1: Signaltest für Ereignis: Alarmeinheit abgeschaltet                  //
//       401F10h:1: Signaltest für Ereignis: Aktivierung der Alarmeinheit angekündigt   //
//       401F20h:1: Signaltest für Ereignis: Alarmeinheit aktiviert                     //
//       401F30h:1: Signaltest für Ereignis: Alarmauslösung angekündigt                 //
//       401F40h:1: Signaltest für Ereignis: Alarm ausgelöst                            //
//                                                                                      //
//  Beschreibung: Löst zu Probezwecken die ausgewählte Alarmeinheit-Ereignis-Signali-   //
//                sierung aus (Ton- und Lichtsignal).                                   //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |TST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//   [TST] Auslöser der Signalisierung bei o.g. Alarmeinheit-Ereignis: 94, 5Eh ('^').   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmEventSignalTest(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 index;
    if     ( location < 0x401F10 ) index = 0;
    else if( location < 0x401F20 ) index = 1;
    else if( location < 0x401F30 ) index = 2;
    else if( location < 0x401F40 ) index = 3;
    else                           index = 4;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E)
    {   EventSignal signal;
        STORING_Load(STO_EventSignal[index], &signal);
        LIGHTSIGNAL_Play(signal.lightSignal, signal.brightness);
        TONESIGNAL_Play(signal.toneSignal);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: AlarmStateSignal (read write)                                               //
//                                                                                      //
//  Beschreibung: Legt die Signalisierung bei bestimmten Zuständen des Alarmsystems     //
//                fest. AlarmStateSignal Gemeinschaftsbearbeiter folgender Dienste:     //
//                                                                                      //
//      402000h:3: Signalisierung im Zustand: Alarmeinheit abgeschaltet                 //
//      402100h:3: Signalisierung im Zustand: Aktivierung der Alarmeinheit angekündigt  //
//      402200h:3: Signalisierung im Zustand: Alarmeinheit aktiviert                    //
//      402300h:3: Signalisierung im Zustand: Alarmauslösung angekündigt                //
//      402400h:3: Signalisierung im Zustand: Alarm ausgelöst                           //
//      402500h:3: Signalisierung bei unbekanntem Zustand                               //
//                                                                                      //
//  Beschreibung: Legt die Art des Licht- und Tonsignals fest, das im jeweiligen Zu-    //
//                stand der Alarmeinheit ausgegeben wird.                               //
//                                                                                      //
//              0   1   2                                                               //
//            +---+---+---+                                                             //
//  Belegung: |TON|LHT|BRT|                                                             //
//            +---+---+---+                                                             //
//                                                                                      //
//  [TON] Tonsignal, wenn Alarmeinheit sich im o.g. Zustand befindet.                   //
//          0 = Lautlos (Voreinstellung)                                                //
//          1 = Tropfen                                                                 //
//          2 = Ticken                                                                  //
//          3 = Martinticks                                                             //
//          4 = Kurzer Piepton D7                                                       //
//          5 = Langer Piepton D7                                                       //
//          6 = Kurzer Piepton D3                                                       //
//          7 = Gelegentlicher Piepton D3 3x                                            //
//          * = Lautlos                                                                 //
//  [LHT] Lichtsignal, wenn Alarmeinheit sich im o.g. Zustand befindet.                 //
//          0 = Aus (Voreinstellung).                                                   //
//          1 = Ein                                                                     //
//          2 = Blinkt                                                                  //
//          3 = Blitzt                                                                  //
//          4 = Pulsiert                                                                //
//          5 = Herzschlag                                                              //
//          6 = Gelegentlicher Blitz 3x                                                 //
//          * = Aus                                                                     //
//  [BRT] Helligkeit der LED bei Lichtsignal [LHT]: 10 .. 100 %                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnAlarmStateSignal(const Cy4Service* service, Cy4Message* message)
{
    UInt32 location = CY4MESSAGE_Location(message);
    UInt08 index;
    if     ( location < 0x402100 ) index = 0;
    else if( location < 0x402200 ) index = 1;
    else if( location < 0x402300 ) index = 2;
    else if( location < 0x402400 ) index = 3;
    else if( location < 0x402500 ) index = 4;
    else                           index = 5;
    StateSignal signal;
    STORING_Load(STO_StateSignal[index], &signal);
    UInt08  data[3];
    SerializeStateSignal(&signal, data);
    CY4SERVICE_TransferPayload(service, message, data);
    signal = DeserializeStateSignal(data);
    STORING_Save(&signal, STO_StateSignal[index]);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: LedBusActivity (read write)                                                 //
//                                                                                      //
//  Beschreibung: Bestimmt, ob und wie hell Bus-Aktivität mit dem Gerät wird auf der    //
//                LED angezeigt wird.                                                   //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ACT|BRT|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [ACT] Bus-Aktivität wird auf der LED angezeigt: 0 = Nein, sonst Ja.                 //
//  [BRT] Helligkeit der LED bei Bus-Aktivität: 10 .. 100 %                             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnBusActivityOnLed(const Cy4Service* service, Cy4Message* message)
{
    LedBusActivity ledBusActivity;
    STORING_Load(STO_LedBusActivity, &ledBusActivity);
    UInt08 data[2];
    data[0] = (bool)ledBusActivity.enabled;
    data[1] = ledBusActivity.brightness;
    CY4SERVICE_TransferPayload(service, message, data);
    ledBusActivity.enabled    = (bool)data[0];
    ledBusActivity.brightness = Bound(10, data[1], 100);
    STORING_Save(&ledBusActivity, STO_LedBusActivity);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N00|N01|N02| ... |N59| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[CARDINALITY(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand.           //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E)
    {   STORING_RestoreDefaults();
        BOARD_Reset();
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: HeartbeatOnSystemLed (read write)                                           //
//                                                                                      //
//  Beschreibung: Bestimmt, ob das Herzschlagsignal auf der System-LED ausgegeben wird. //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |ACT|BRT|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [HBT] Herzschlagsignal auf der System-LED: 0 = Aus, sonst Ein.                      //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnHeartbeatOnSystemLed(const Cy4Service* service, Cy4Message* message)
{
    bool  heartbeatEnabled;
    STORING_Load(STO_HeartbeatOnSystemLed, &heartbeatEnabled);
    UInt08 data = (UInt08)heartbeatEnabled;
    CY4SERVICE_TransferPayload(service, message, &data);
    heartbeatEnabled = (bool)data;
    STORING_Save(&heartbeatEnabled, STO_HeartbeatOnSystemLed);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^') .                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^') .                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeDeviceAddress (write only)                                     //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeDeviceAddress(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Verfahrensfunktionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedContacting(): Verfahrensfunktion der Kontaktierung. Die Zustand des Auto-    //
//  -------------------  maten findet sich in <contactState>.                           //
//                                                                                      //
//   Zustandsübergangsdiagramm:                                                         //
//                                                   IS_CLOSED                          //
//                               +-----------------+ -> StartTimer(a)                   //
//              +--------------->|       OPEN      +-------------+                      //
//              |                +-----------------+             |                      //
//              |                        /|\  /|\                |                      //
//              |                         |    |                \|/                     //
//              |                         |    |   IS_OPEN +-----------+                //
//              |                         |    +-----------+  CLOSING  |                //
//              |                         |                +-----+-----+                //
//              |                         | IS_OPEN              | TIMEOUT              //
//              |                         | -> SetState(open)    | -> SetState(closed)  //
//              |                        (O)                     |                      //
//              | TIMEOUT                 | IS_CLOSED            |                      //
//              | -> SetState(open)       | -> SetState(closed)  |                      //
//        +-----+-----+                   |                      |                      //
//        |  OPENING  +--------------+    |                      |                      //
//        +-----------+ IS_CLOSED    |    |                      |                      //
//             /|\                   |    |                      |                      //
//              |                   \|/  \|/                     |                      //
//              |                +-----------------+             |                      //
//              +----------------|     CLOSED      |<------------+                      //
//                   IS_OPEN     +-----------------+                                    //
//                   -> StartTimer(b)                                                   //
//                                                                                      //
//      a = Zustandsdauer bis Kontaktereignis "gesschlossen" eintritt                   //
//      b = Zustandsdauer bis Kontaktereignis "geöffnet" eintritt                       //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedContacting(void)
{
    typedef enum     // Kontaktierungs-Zustände:
    {   OPEN,        // · Kontakt geöffnet
        CLOSING,     // · Kontakt schliessend
        CLOSED,      // · Kontakt geschlossen
        OPENING,     // · Kontakt öffnend
        UNKNOWN      // · Kontaktzustand unbekannt (Anfangszustand)
    } State;

    static Timer msTimer[CFG_CONTACT_COUNT]  = { TIMER_MILLISECONDS(0),
                                                 TIMER_MILLISECONDS(0),
                                                 TIMER_MILLISECONDS(0)
                                               };
    static Timer secTimer[CFG_CONTACT_COUNT] = { TIMER_SECONDS(0),
                                                 TIMER_SECONDS(0),
                                                 TIMER_SECONDS(0)
                                               };
    static State state[CFG_CONTACT_COUNT]    = { UNKNOWN, UNKNOWN, UNKNOWN };
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   bool isClosed = BOARD_IsContactClosed(i);
        switch (state[i])
        {   case OPEN:
                if (isClosed)
                {   UInt08 duration;
                    STORING_Load(STO_ContactEventConfiguration[i][0].stateDurationUntilEvent, &duration);
                    if      (duration ==  0) TIMER_Start(&msTimer[i] , CFG_CONTACT_DEBOUNCE_TIME);
                    else if (duration <  15) TIMER_Start(&msTimer[i] , (UInt16)duration*1000);
                    else                     TIMER_Start(&secTimer[i], duration);
                    state[i] = CLOSING;
                }
                break;
            case CLOSING:
                if (!isClosed)
                {   state[i] = OPEN;
                }
                else if (TIMER_HasExpired(&msTimer[i]) && TIMER_HasExpired(&secTimer[i]))
                {   state[i] = CLOSED;
                    contactState[i] = CONTACT_STATE_CLOSED;
                }
                break;
            case CLOSED:
                if (!isClosed)
                {   UInt08 duration;
                    STORING_Load(STO_ContactEventConfiguration[i][1].stateDurationUntilEvent, &duration);
                    if      (duration ==  0) TIMER_Start(&msTimer[i] , CFG_CONTACT_DEBOUNCE_TIME);
                    else if (duration <  15) TIMER_Start(&msTimer[i] , (UInt16)duration*1000);
                    else                     TIMER_Start(&secTimer[i], duration);
                    state[i] = OPENING;
                }
                break;
            case OPENING:
                if (isClosed)
                {   state[i] = CLOSED;
                }
                else if (TIMER_HasExpired(&msTimer[i]) && TIMER_HasExpired(&secTimer[i]))
                {   state[i] = OPEN;
                    contactState[i] = CONTACT_STATE_OPEN;
                }
                break;
            default:
                // UNKNOWN
                if (!isClosed)
                {   state[i] = OPEN;
                    contactState[i] = CONTACT_STATE_OPEN;
                }
                else
                {   state[i] = CLOSED;
                    contactState[i] = CONTACT_STATE_CLOSED;
                }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedContactSignaling(void)
{
    static ContactState previousContactState[CFG_CONTACT_COUNT] = { CONTACT_STATE_UNKNOWN,
                                                                    CONTACT_STATE_UNKNOWN,
                                                                    CONTACT_STATE_UNKNOWN
                                                                  };

    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   if (previousContactState[i] != contactState[i])
        {   if (previousContactState[i] != CONTACT_STATE_UNKNOWN)
            {   ContactAlarmRole closeEventRole;
                STORING_Load(STO_ContactEventConfiguration[i][0].alarmRole, &closeEventRole);
                ContactAlarmRole openEventRole;
                STORING_Load(STO_ContactEventConfiguration[i][1].alarmRole, &openEventRole);
                // Signal nur ausgeben, wenn eine der folgenden drei Bedingungen erfüllt ist:
                //  (1) Kontaktereignisse im Alarmsystem haben keine Aufgabe
                //  (2) Das Alarmsystem ist im unbekanntem Zustand
                //  (3) Kontaktereignisse einen Alarm auslösen aber das Alarmsystem abgeschaltet ist
                if ((closeEventRole == CONTACT_ALARM_ROLE_NONE && openEventRole == CONTACT_ALARM_ROLE_NONE) ||      // (1)
                    ALARMUNIT_State() == ALARMUNIT_STATE_UNKNOWN ||                                                 // (2)
                    (ALARMUNIT_State() == ALARMUNIT_STATE_DISABLED && (closeEventRole == CONTACT_ALARM_ROLE_TRIGGER || openEventRole == CONTACT_ALARM_ROLE_TRIGGER)))  // (3)
                {   UInt08 openEvent = 0xFF;
                    switch (contactState[i])
                    {   case CONTACT_STATE_OPEN:   openEvent = 1; break;
                        case CONTACT_STATE_CLOSED: openEvent = 0; break;
                        default:                   ;
                    }
                    if (openEvent < 2)
                    {   EventSignal signal;
                        STORING_Load(STO_ContactEventConfiguration[i][openEvent].eventSignal, &signal);
                        LIGHTSIGNAL_Play(signal.lightSignal, signal.brightness);
                        TONESIGNAL_Play(signal.toneSignal);
                    }
                }
            }
            previousContactState[i] = contactState[i];
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedAlarming(void)
{
    static ContactState previousContactState[CFG_CONTACT_COUNT] = { CONTACT_STATE_UNKNOWN,
                                                                    CONTACT_STATE_UNKNOWN,
                                                                    CONTACT_STATE_UNKNOWN
                                                                  };
    for (UInt08 i=0; i<CFG_CONTACT_COUNT; i++)
    {   if (previousContactState[i] != contactState[i])
        {   ContactAlarmRole alarmRole;
            switch (previousContactState[i])
            {   case CONTACT_STATE_OPEN:
                    STORING_Load(STO_ContactEventConfiguration[i][0].alarmRole, &alarmRole);
                    break;
                case CONTACT_STATE_CLOSED:
                    STORING_Load(STO_ContactEventConfiguration[i][1].alarmRole, &alarmRole);
                    break;
                default:
                    alarmRole = CONTACT_ALARM_ROLE_NONE;
            }
            previousContactState[i] = contactState[i];
            switch (alarmRole)
            {   case CONTACT_ALARM_ROLE_ACTIVATE:
                {   UInt08 duration;
                    STORING_Load(STO_ActivationAnnouncementDuration, &duration);
                    ALARMUNIT_Activate(duration);
                    break;
                }
                case CONTACT_ALARM_ROLE_DEACTIVATE:
                {   ALARMUNIT_Deactivate();
                    break;
                }
                case CONTACT_ALARM_ROLE_TOGGLE:
                {   UInt08 duration;
                    STORING_Load(STO_ActivationAnnouncementDuration, &duration);
                    ALARMUNIT_Toggle(duration);
                    break;
                }
                case CONTACT_ALARM_ROLE_TRIGGER:
                {   UInt08 duration;
                    STORING_Load(STO_TriggerAnnouncementDuration, &duration);
                    const char* text = STORING_AddressOf(STO_TriggerTextFeed.text);
                    ALARMUNIT_Trigger(duration, text);
                    break;
                }
                default:
                ;
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedAlarmSignaling(void)
{
    static AlarmUnitState prevState = ALARMUNIT_STATE_TRIGGERED;

    AlarmUnitState state = ALARMUNIT_State();
    if (state == ALARMUNIT_STATE_STARTED) return;
    UInt08 index = (UInt08)state-1;
    // Ereignis-Signale
    if (prevState != state)
    {   prevState  = state;
        // Tonausgabe
        {   ToneEventSignal toneSignal;
            STORING_Load(STO_EventSignal[index].toneSignal, &toneSignal);
            TONESIGNAL_Play(toneSignal);
        }
        // Lichtausgabe
        {   LightEventSignal lightSignal;
            STORING_Load(STO_EventSignal[index].lightSignal, &lightSignal);
            UInt08 brightness;
            STORING_Load(STO_EventSignal[index].brightness, &brightness);
            LIGHTSIGNAL_Play(lightSignal, brightness);
        }
    }
    // Zustandssignale
    {   // Tonausgabe
        {   ToneStateSignal toneSignal;
            STORING_Load(STO_StateSignal[index].toneSignal, &toneSignal);
            if (TONESIGNAL_Current() != toneSignal) TONESIGNAL_Start(toneSignal);
        }
        // Lichtausgabe
        {   LightStateSignal lightSignal;
            UInt08           brightness;
            STORING_Load(STO_StateSignal[index].lightSignal, &lightSignal);
            STORING_Load(STO_StateSignal[index].brightness , &brightness);
            if (LIGHTSIGNAL_Current() != lightSignal || LIGHTSIGNAL_Brightness() != brightness)
            {   LIGHTSIGNAL_Start(lightSignal, brightness);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
ProceedSystemLed(void)
{
    bool ledOn = false;
    {   bool  heartbeatEnabled;
        STORING_Load(STO_HeartbeatOnSystemLed, &heartbeatEnabled);
        if (heartbeatEnabled)
        {   UInt16 ticks = BOARD_MSecTicks();
            ledOn = (bool)((ticks & 0x380)==0x80 || (ticks & 0x380)==0x180);
        }
    }
    if (BOARD_BusActivity()) ledOn = true;
    BOARD_SetSystemLed(ledOn);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                     Dienst                     |          Bearbeiter          |
        //+------------------------------------------------+------------------------------+
        { CY4SERVICE(0x130000,  3  , CY4SERVICE_READ_ONLY) , OnContactState               },
        { CY4SERVICE(0x130100,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x13017A,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        { CY4SERVICE(0x130180,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x1301FA,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        { CY4SERVICE(0x130200,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x13027A,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        { CY4SERVICE(0x130280,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x1302FA,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        { CY4SERVICE(0x130300,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x13037A,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        { CY4SERVICE(0x130380,  5  , CY4SERVICE_READ_WRITE), OnContactEventConfiguration  },
        { CY4SERVICE(0x1303FA,  1  , CY4SERVICE_WRITE_ONLY), OnContactEventSignalTest     },
        // Dienste der Alarmeinheit
        { CY4SERVICE(0x200000,  2  , CY4SERVICE_READ_WRITE), OnAlarmAnnouncementDuration  },
        { CY4SERVICE(0x200100, 51  , CY4SERVICE_READ_WRITE), OnAlarmTriggerTextFeed       },
        { CY4SERVICE(0x400100,  1  , CY4SERVICE_READ_ONLY) , OnAlarmUnitState             },
        { CY4SERVICE(0x401000,  3  , CY4SERVICE_READ_WRITE), OnAlarmEventSignal           },
        { CY4SERVICE(0x401100,  3  , CY4SERVICE_READ_WRITE), OnAlarmEventSignal           },
        { CY4SERVICE(0x401200,  3  , CY4SERVICE_READ_WRITE), OnAlarmEventSignal           },
        { CY4SERVICE(0x401300,  3  , CY4SERVICE_READ_WRITE), OnAlarmEventSignal           },
        { CY4SERVICE(0x401400,  3  , CY4SERVICE_READ_WRITE), OnAlarmEventSignal           },
        { CY4SERVICE(0x401F00,  1  , CY4SERVICE_WRITE_ONLY), OnAlarmEventSignalTest       },
        { CY4SERVICE(0x401F10,  1  , CY4SERVICE_WRITE_ONLY), OnAlarmEventSignalTest       },
        { CY4SERVICE(0x401F20,  1  , CY4SERVICE_WRITE_ONLY), OnAlarmEventSignalTest       },
        { CY4SERVICE(0x401F30,  1  , CY4SERVICE_WRITE_ONLY), OnAlarmEventSignalTest       },
        { CY4SERVICE(0x401F40,  1  , CY4SERVICE_WRITE_ONLY), OnAlarmEventSignalTest       },
        { CY4SERVICE(0x402000,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        { CY4SERVICE(0x402100,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        { CY4SERVICE(0x402200,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        { CY4SERVICE(0x402300,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        { CY4SERVICE(0x402400,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        { CY4SERVICE(0x402500,  3  , CY4SERVICE_READ_WRITE), OnAlarmStateSignal           },
        // Rundruf-Dienste des Alarmsystems (Rundrufzonenbereich: E020xxh)
        { CY4SERVICE(0xE02000,  4  , CY4SERVICE_WRITE_ONLY), OnALARMUNIT_StateBroadcast   },
        // Allgemeine Gerätedienste
        { CY4SERVICE(0xFFF020,  2  , CY4SERVICE_READ_WRITE), OnBusActivityOnLed           },
        { CY4SERVICE(0xFFFF00, 61  , CY4SERVICE_READ_WRITE), OnDeviceLabel                },
        { CY4SERVICE(0xFFFF40,  1  , CY4SERVICE_WRITE_ONLY), OnFactorySettings            },
        { CY4SERVICE(0xFFFF50,  1  , CY4SERVICE_READ_WRITE), OnHeartbeatOnSystemLed       },
        // Pflichtdienste
        { CY4SERVICE(0xFFFFF0,  7  , CY4SERVICE_READ_ONLY ), OnCallTypePlate              },
        { CY4SERVICE(0xFFFFF8,  1  , CY4SERVICE_WRITE_ONLY), OnEnterApplication           },
        { CY4SERVICE(0xFFFFFA,  1  , CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader         },
        { CY4SERVICE(0xFFFFFC,  2  , CY4SERVICE_WRITE_ONLY), OnChangeDeviceAddress        }
    };

    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    ProceedContacting();
    ProceedContactSignaling();
    ProceedAlarming();
    ProceedAlarmSignaling();
    ProceedSystemLed();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
